--
-- Cabsuspension
-- Specialization for Cabsuspension
--
-- @author  Spartan086
-- @date  2017-03-21
-- @version 1.0
-- Free to use but no support given


Cabsuspension = {};

function Cabsuspension.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Motorized, specializations);
end;

function Cabsuspension:load(savegame)
	--Create base object
	self.CS = {};
	self.CS.enabled = true;
	
	--Read from xml file
	local base = "vehicle.cabSuspension";
	self.CS.baseNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, base .. "#baseNode"));
	if self.CS.baseNode == nil then
		--print("failed to find baseNode");
		self.CS.enabled = false;
	end 
	self.CS.cabNode =  Utils.indexToObject(self.components, getXMLString(self.xmlFile, base .. "#cabNode"));
	if self.CS.cabNode == nil then
		--print("failed to find cabNode");
		self.CS.enabled = false;
	end 
	
	self.CS.wLIdx = tonumber( Utils.getNoNil(getXMLString(self.xmlFile, base .. "#leftWheelIndex"), 0));
	self.CS.wRIdx = tonumber( Utils.getNoNil(getXMLString(self.xmlFile, base .. "#rightWheelIndex"), 1));
	
	self.CS.forceFactorX = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".pitch#forceFactor"), 50)) / 1000;
	self.CS.centeringFactorX = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".pitch#centeringFactor"), 30)) / 1000;
	self.CS.thresholdX = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".pitch#threshold"), 5)) / 100;
	self.CS.maxRotX = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".pitch#maxRot"), 1.0));
	if self.CS.maxRotX == 0 then
		self.CS.enabled = false; --prevent division by 0
	end;
	
	self.CS.forceFactorZ = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".roll#forceFactor"), 50)) / 1000;
	self.CS.centeringFactorZ = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".roll#centeringFactor"), 30)) / 1000;
	self.CS.thresholdZ = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".roll#threshold"), 5)) / 100;
	self.CS.maxRotZ = tonumber(Utils.getNoNil(getXMLFloat(self.xmlFile, base .. ".roll#maxRot"), 3.0));
	if self.CS.maxRotZ == 0 then
		self.CS.enabled = false; --prevent division by 0
	end;
end;

function Cabsuspension:postLoad(savegame)
	--Initialize positions, rotations etc.
	--print("Init cab suspension");
	local wheel = self.wheels[self.CS.wLIdx];
	self.CS.wL = wheel.repr;
	wheel = self.wheels[self.CS.wRIdx];
	self.CS.wR = wheel.repr;
	self.CS.realTravel = wheel.suspTravel;
	if self.CS.realTravel == 0 then
		self.CS.enabled = false; --Prevent division by 0
	end;
	
	--Fetch rest-positions
	local lx, ly, lz = getTranslation(self.CS.wL);
	self.CS.restL = ly;
	local rx, ry, rz = getTranslation(self.CS.wR);
	self.CS.restR = ry;
	
	
	--Calc distance between left / right
	--self.CS.distance = math.abs(lx - rx);
	self.CS.lastSpeed = 0;
	
	self.CS.currX = 0;
	self.CS.currZ = 0;
	
	--print("Rest L : ".. lx .. " " .. ly .. " " .. lz);
	--print("Rest R : ".. rx .. " " .. ry .. " " .. rz);
	--print("distance: " .. self.CS.distance);
end


function Cabsuspension:delete()
end;

function Cabsuspension:readStream(streamId, connection)
end;

function Cabsuspension:writeStream(streamId, connection)
end;

function Cabsuspension:readUpdateStream(streamId, timestamp, connection)
end;

function Cabsuspension:writeUpdateStream(streamId, connection, dirtyMask)
end;

function Cabsuspension:mouseEvent(posX, posY, isDown, isUp, button)
end;

function Cabsuspension:keyEvent(unicode, sym, modifier, isDown)
end;


function Cabsuspension:update(dt)
	--Quick disabling of updates
	if not self.CS.enabled then
		return 0;
	end;
	
	if self.isClient then
		if self:getIsActive() then
			Cabsuspension.updateRotation(self, dt);
		end
	end
end

function Cabsuspension.updateRotation(self, dt)
	local lx, ly, lz = getTranslation(self.CS.wL);
	self.CS.posL = self.CS.restL - ly;
	
	local rx, ry, rz = getTranslation(self.CS.wR);
	self.CS.posR = self.CS.restR - ry;
	
    local speed = self:getLastSpeed();
	--local speedFactor = Utils.clamp(speed, -5, 5) / 5;
	local speedFactor = 1;
	
	
	local hDiff = (self.CS.posL - self.CS.posR) / (-self.CS.realTravel/10); --Percent of max difference
	hDiff = Utils.clamp(hDiff, -1, 1);
	local centerDir = math.abs(self.CS.currZ) / -self.CS.currZ; --inverse or current rotation
	if centerDir ~= centerDir then -- check if 0/0 (NaN)
		centerDir = 0;
	end;
	
	local force = hDiff * self.CS.forceFactorZ * dt * speedFactor;
	--force = Utils.clamp(force, -1, 1);
	-- if math.abs(force) < 0.05 then --if the force is small and we are close to 0, ignore it!
		-- force = 0;
	-- end;
	
	local centering = (math.abs(self.CS.currZ)/self.CS.maxRotZ) * self.CS.centeringFactorZ * dt * centerDir;
	-- if math.abs(centering) < 0.05 then
		-- centering = 0;
	-- end;

	--print("force " .. round(force, 4) .. " center: " .. round(centering, 4) .. " currZ" .. round(self.CS.currZ, 4));
	if math.abs(force + centering) < self.CS.thresholdZ then
		--print("Ignore minor forces Z " .. (force + centering));
		force = 0;
		centering = 0;
	end;
	local rotation = force + centering;
	rotation = Utils.clamp(rotation, -0.0025*dt, 0.0025*dt); 
	self.CS.currZ = self.CS.currZ + rotation; --i.e if hDiff > 0.25 then the cab will move
	self.CS.currZ = Utils.clamp(self.CS.currZ, -self.CS.maxRotZ, self.CS.maxRotZ);
	
	
	
	
	--Tilt front/Back
	local speedDiff = (self.CS.lastSpeed - speed); --speed diff in m/s 
	--local speedPart = speedDiff / dt
	--local transPart = ((self.CS.posL + self.CS.posR) / 2) / (self.CS.realTravel);
	--print("speedPart " .. round(speedPart, 4) .. " transPart " .. round(transPart, 4) .. " cDiff " .. round(cDiff, 4) .. " vs " .. round();
	--local cDiff = speedPart + transPart;
	--cDiff = Utils.clamp(speedPart, -1, 1) * self.movingDirection;
	--local cDiff = ((self.CS.posL + self.CS.posR) / 2) / (self.CS.realTravel);
	centerDir = math.abs(self.CS.currX) / -self.CS.currX; --inverse or current rotation
	if centerDir ~= centerDir then -- check if 0/0 (NaN)
		centerDir = 0;
	end;
	--print("speedPart " .. round(speedPart, 4) .. " transPart " .. round(transPart, 4) .. " cDiff " .. round(cDiff, 4));
	
	force = speedDiff * self.movingDirection * self.CS.forceFactorX * dt;
	--local before = hDiff;
	--force = Utils.clamp(force, -1, 1);
	--print("Before: " .. round(before, 4) .. " after: " .. round(hDiff, 4));
	centering = (0.05+math.abs(self.CS.currX)/self.CS.maxRotX) * self.CS.centeringFactorX * dt * centerDir;
	--print("force " .. round(force, 4) .. " center: " .. round(centering, 4) .. " currX" .. round(self.CS.currX, 4));
	
	-- if math.abs(force) < 0.05 then --if the force is small and we are close to 0, ignore it!
		-- --print("force 0");
		-- force = 0;
	-- end;
	
	-- if math.abs(centering) < 0.1 then
		-- --print("center 0");
		-- centering = 0;
	-- end;
	
	if math.abs(force + centering) < self.CS.thresholdX then
		--print("Ignore minor forces X " .. (force + centering));
		force = 0;
		centering = 0;
	end;
	
	rotation = force + centering;
	rotation = Utils.clamp(rotation, -0.0025*dt, 0.0025*dt); 
	self.CS.currX = self.CS.currX + rotation; 
	self.CS.currX = Utils.clamp(self.CS.currX, -self.CS.maxRotX, self.CS.maxRotX);
	
	
	self.CS.lastSpeed = speed;
	setRotation(self.CS.cabNode, math.rad(self.CS.currX), 0, math.rad(self.CS.currZ));
end

function round(val, decimal)
  if (decimal) then
    return math.floor( (val * 10^decimal) + 0.5) / (10^decimal)
  else
    return math.floor(val+0.5)
  end
end

function Cabsuspension:updateTick(dt)

end;

function Cabsuspension:draw()
end;

function Cabsuspension:onLeave()

end;

function Cabsuspension:onEnter()

end;
